# !/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Module that runs when you natively execute the program.

As if one did this:
>>> python3 -m quizmake
"""

import logging
import os
import re
from typing import List

from quizmake import parser


def main(argv: List[str]) -> int:
    """
    Take in the user's question queries and process them.

    :param argv: The sys.argv list of command-line args
    :type argv: List[str]
    :return: The exit code of the program, usually.
    :rtype: int
    """
    try:
        args = parser.verify_args(argv[1:])
    except ValueError:
        return 1

    # Set up logging
    logging.basicConfig(
        level=logging.DEBUG if args.verbose else logging.INFO,
        format="[%(levelname)s] %(message)s",
    )

    logging.info("Logging initiated")

    """
    Not only do both `tokens` and `questions` must exist as directories
    have to exist as directories: they must also have a nonzero amount
    of valid files (question/answer, tokens, etc.)

    In the original script that I made, it would only check for existence
    of the directories but never the validity of the files themselves. This
    meant that the program would not realize this until much later in the
    export of the resultant files. Here I am to save some time.
    """

    t_folder = args.tokens
    q_folder = args.questions

    logging.info(f'Checking path "{t_folder}" for tokens')
    if not parser.assert_dir_has_files(t_folder):
        logging.error(f"{t_folder} either has no tokens or doesn't exist")
        logging.error("Exiting...")
        return 1

    logging.info(f'Checking path "{q_folder}" for questions')
    if not parser.assert_dir_has_files(q_folder):
        logging.error(f"{q_folder} either has no questions or doesn't exist")
        logging.error("Exiting...")
        return 1

    """
    This is going to be a relatively slow (but important) step where each and
    every token and question file will be validated and parsed into objects.

    For tokens, it's going to be relatively quick, as tokens are delimited
    merely line-by-line.

    For questions, on the other hand... :grimace:

    TODO: in this part of the project, I haven't thought of the top-down
          result of parsing tokens and questions. I just want to parse.

          In the original project, I had a 'DataSet' object that basically
          represented each file in the tokens folder. Then, I had a 'Corpus'
          object that held a dictionary of 'DataSet' objects and a top-level
          interface was made from that. I might do the same, but less messy.

          The original implementation used a very weird index system.
    """

    # PLANNED: integrate validation and creation into one step
    for token in os.listdir(t_folder):
        logging.debug(f"Testing token {token}")
        if not parser.assert_token_file(t_folder + "/" + token):
            raise Exception(f"{token} is not a valid token.")
            # logging.error(f"{token} is not a valid token. Skipping...")
        else:
            logging.info(f"{token} is valid...")

    logging.info(f"Tokens folder {t_folder} is valid. Parsing files now.")
    try:
        corpus = parser.Corpus(t_folder)
    except Exception as e:
        logging.critical(f"Could not parse tokens folder {t_folder}.")
        logging.critical(f"Exception: {e}")
        exit(1)

    questions_array = []
    filenames_array = []
    for question in os.listdir(q_folder):
        logging.debug(f"Testing question {question}")
        parser.assert_question_file(q_folder + "/" + question)
        if not parser.assert_question_file(q_folder + "/" + question):
            raise Exception(f"{question} is not a valid questiion.")
            # logging.error(f"{question} is not a valid question. Skipping...")
        else:
            logging.info(f"{question} is valid...")
            temp = parser.Question(q_folder + "/" + question)
            questions_array.append(temp)
            filenames_array.append(question)

    class bcolors:
        HEADER = '\033[95m'
        OKBLUE = '\033[94m'
        OKGREEN = '\033[92m'
        WARNING = '\033[93m'
        FAIL = '\033[91m'
        ENDC = '\033[0m'
        BOLD = '\033[1m'
        UNDERLINE = '\033[4m'

    def print_question(filename, question):
        question_text = question.sections["question"][0]
        answers = question.sections["answers"]
        feedback = question.sections["feedback"]

        print("="*50)
        print(f"{filename}: ", end = "")
        print(question_text + "\n")

        option_number = 1
        print("")
        for i in range(len(answers)):
            print(f"{option_number}. {answers[i]} ", end = "")
            print(bcolors.OKGREEN + "(correct)" + bcolors.ENDC if i == 0 else "", end = "")
            print(f"\n(Feedback: {feedback[i] if feedback[i].rstrip() else 'None'})")
            option_number += 1
            print()

    identifier = re.compile(r'{([A-Za-z-_]+)([0-9]+)?}')
    custom_call = re.compile(r'{CUSTOM\.([\w\_]+)}')

    def process_line(corpus, line):
        matches = identifier.findall(line)
        for match in matches:
            filename = match[0]
            number = match[1] if match[1] else ''
            original_text = '{' + filename + number + '}'
            if corpus.exists(filename):
                request = corpus.request(filename, int(number if number else 0))
                line = line.replace(original_text, request, 1)
        return line
        

    for question in questions_array:
        corpus = parser.Corpus(t_folder) # TODO: proper corpus reset
        sections = [
            "question",
            "answers",
            "feedback",
        ]
        for section in sections:
            for i in range(len(question.sections[section])):
                question.sections[section][i] = process_line(corpus, question.sections[section][i])

    def quick_clean(dirty_string):
        return dirty_string\
            .replace("\\n", "\\\\n")\
            .replace("~", "\\~")\
            .replace("=", "\\=")\
            .replace("#", "\\#")\
            .replace("{", "\\{")\
            .replace("}", "\\}")\
            .replace(":", "\\:")\
            .replace("<", "&lt;")\
            .replace(">", "&gt;")\
            .replace("    ", "&nbsp;&nbsp;&nbsp;&nbsp;")\
            .replace("\n", "<br>")

    def output_gift(question_number, question, filename):
        output = open(filename, "a")
        text = question.sections["question"]
        answers = question.sections["answers"]
        feedbacks = question.sections["feedback"]
        output.write(f'::Q{question_number}::[html]')
        text_cleaned = quick_clean(text[0])
        for i in range(len(answers)):
            answers[i] = quick_clean(answers[i])
            feedbacks[i] = quick_clean(feedbacks[i])
        output.write(text_cleaned + "\n")
        output.write('{ =')
        for i in range(len(answers[:-1])):
            output.write(answers[i])
            output.write(' # ')
            output.write(feedbacks[i])
            output.write(' ~')
        output.write(answers[- 1])
        output.write(' # ' + feedbacks[-1])
        output.write('}\n\n')

    if args.export_gift:
        logging.info(f"Exporting GIFT to {args.export_gift}!")
        for i in range(len(questions_array)):
            output_gift(i + 1, questions_array[i], args.export_gift)
        logging.info("Done!")
    else:
        counter = 0
        for question in questions_array:
            print_question(filenames_array[counter], question)
            counter += 1

    return 0





















# !/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Initiate the file."""
__author__ = "Jason Nguyen"
__copyright__ = "Copyright 2020, Jason Nguyen"
__credits__ = ["Jason Nguyen"]
__license__ = "MIT"
__maintainer__ = "Jason Nguyen"
__email__ = "jnguye21@uoguelph.ca"
__status__ = "Planning"
__version__ = "0.0.8.3"

# from .core import main
# from .corpus import Corpus, corn
# __all__ = ["Corpus", "corn", "main"]



# !/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Common helper functions.

This will represent the bread and butter of the application.
"""

import argparse
import copy
import logging
import os
import random
import sys
from enum import Enum
from typing import List, NoReturn

from pyparsing import ParseException

from quizmake import grammar


class QuestionType(Enum):
    """Enumeration of the question types."""

    OTHER = 0
    MULTIPLE_CHOICE = 1
    SHORT_ANSWER = 2
    TRUE_FALSE = 3
    MATCHING = 4
    NUMERICAL = 5
    ESSAY = 6
    DESCRIPTION = 7
    CLOZE = 8


# Question Object
class Question:
    """Used to store question file data."""

    def __init__(self, filename: str) -> None:
        """Create a question using the filename."""
        self.sections = {}
        self.filename = filename
        self.parsed = grammar.parse_file(filename)

        for section in self.parsed:

            if section[0].casefold() == "[multiple_choice]".casefold():
                self.type = QuestionType.MULTIPLE_CHOICE
                question_lines: List[str] = []
                for line in section[1:]:
                    if not line.startswith("//"):
                        question_lines.append(line)
                self.sections["question"] = ["\n".join(question_lines)]

            elif section[0].casefold() == "[answer]".casefold():
                answer_lines: List[str] = []
                for line in section[1:]:
                    if not line.startswith("//"):
                        answer_lines.append(line)
                self.sections["answers"] = answer_lines

            elif section[0].casefold() == "[feedback]".casefold():
                feedback_lines: List[str] = []
                for line in section[1:]:
                    if not line.startswith("//"):
                        feedback_lines.append(line)
                self.sections["feedback"] = feedback_lines
            # TODO: re-add else statement
                

    def speak(self) -> None:
        """Thwart the linter lmao."""
        print(self.sections)

    def speaker(self) -> None:
        """Thwart the linter again."""
        print(self.filename)


# Token database
class TokenSet:
    """Used to store token data."""

    def __init__(self, filename: str) -> None:
        """Create a token list using the filename."""

    def __init__(self, filename):
        with open(filename, "r") as fd:
            self.data = fd.read().splitlines()
            self.data = list(filter(lambda a : a, self.data))
            self.uniques = random.sample(copy.deepcopy(self.data), len(self.data))

    def get_random(self):
        return random.choice(self.data)

    def pop_random(self):
        if self.uniques:
            elem = self.uniques.pop()
            return elem
        return None

    def reset_uniques(self):
        self.uniques = copy.deepcopy(self.data)
        random.shuffle(self.uniques)


# Corpus
class Corpus:
    """Used to store a bunch of tokens and their indexes."""

    # TODO: this is bad
    def __init__(self, tokens_folder):
        self.data = {}
        for filename in os.listdir(tokens_folder):
            self.data[filename] = { 0 : TokenSet(tokens_folder + "/" + filename) }

    def exists(self, filename):
        return filename in self.data

    def request(self, filename, number):
        if not self.exists(filename):
            return None # should raise exception
        if number == 0:
            return self.data[filename][0].get_random()
        if number not in self.data[filename]:
            elem = self.data[filename][0].pop_random()
            if not elem:
                raise Exception(f"No unique strings left in {filename}")
            self.data[filename][number] = elem
        return self.data[filename][number]

    def items(self):
        return self.data.items()


class CaughtArgumentParser(argparse.ArgumentParser):
    """Overridden argparse.ArgumentParser class.

    The original ArgumentParser, when encountering an error, would
    completely exit the program. For my testing purposes (and due
    to the fact that I am still very new to PyTest), I ended up
    intercepting the parse_args() call to raise an exception instead
    of outright quitting the program with sys.exit()
    """

    def error(self, message: str) -> NoReturn:
        """Override argparse.ArgumentParser error to not exit."""
        self.print_usage(sys.stderr)
        args = {"prog": self.prog, "message": message}
        sys.stderr.write("%(prog)s: error: %(message)s\n" % args)
        raise ValueError()


def verify_args(argv: List[str]) -> argparse.Namespace:
    """
    Verify the caller's arguments w.r.t. quizmake.

    :param argv: The sys.argv list of command-line args
    :type argv: List[str]
    :return: Dictionary of command-line arguments
    :rtype: Dict[Str, Str]
    """
    arg_p = CaughtArgumentParser()

    arg_p.add_argument("tokens", help="specify tokens folder")
    arg_p.add_argument("questions", help="specify questions folder")
    arg_p.add_argument(
        "--export-gift",
        help="specify GIFT export and output filename",
        metavar="FILENAME",
    )
    arg_p.add_argument(
        "-v", "--verbose", help="verbose debug output", action="store_true"
    )

    return arg_p.parse_args(argv)


def assert_nonempty_dir(folder: str) -> bool:
    """
    Assert that a directory is not empty.

    This means subfolders count!

    :param folder: the folder
    :type folder: str
    :return: Truth statement
    :rtype: bool
    """
    if not os.path.isdir(folder) or not os.listdir(folder):
        return False
    return True


def assert_dir_has_files(folder: str) -> bool:
    """
    Assert that a directory has files.

    This differs from the assert_nonempty_dir function
    because a folder with no files but only subfolders
    will return true when passed through that function.

    :param folder: The folder
    :type folder: str
    :return: Truth statement
    :rtype: bool
    """
    if not assert_nonempty_dir(folder):
        return False

    count = 0
    for _ in os.listdir(folder):
        if os.path.isfile(folder + "/" + _):
            count += 1

    return count > 0


def assert_token_file(filename: str) -> bool:
    """
    Check if a token file has correct structure.

    :param filename: File path to the token file
    :type filename: str:
    :return: Truth statement answering the question
    :rtype: bool
    """
    try:
        with open(filename, "r") as token_fp:
            i = 0
            for i, dummy in enumerate(token_fp, 1):
                pass
        return i > 0
    except (OSError, IOError) as exception:
        logging.debug(f"Token file exception: {exception}")
        return False


def assert_question_file(filename: str) -> bool:
    """
    Check if a question file has correct structure.

    :param filename: File path to the question file
    :type filename: str
    :return: Truth statement answering the question
    :rtype: bool
    """
    try:
        # Question(filename)
        list(grammar.parse_file(filename))
        return True
    except ParseException as exception:
        logging.debug(f"Question file exception: {exception}")
        return False



# !/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Common helper functions.

This will represent the bread and butter of the application.
"""


def add_two_numbers(x_val: int, y_val: int) -> int:
    """Just adds two numbers together."""
    return x_val + y_val


def return_string() -> str:
    """Return the string doggo."""
    return "string"



# !/usr/bin/env python3
# -*- coding: utf-8 -*-


"""
Corpus module used for storing tokens' data.

I might just deprecate this.
"""

__license__ = "MIT"
__revision__ = "not_defined"
__docformat__ = "reStructuredText"


class Corpus:
    """This is just the Corpus class to hold datasets.

    There is a function :func:`speak` that just speaks the
    field `amazing`.

    :param nothing: description I guess
    :param nothing2: description again I guess
    :type nothing: int
    :type nothing2: str
    :return: everything
    :rtype: int
    """

    def __init__(self) -> None:
        """Initiate the Corpus class."""
        self.amazing = "lmao"

    def speak(self) -> None:
        """Speak the assigned word."""
        print(self.amazing)

    def yell(self) -> None:
        """Say random stuff."""
        print(self.amazing + " LMaO")


def corn() -> str:
    """Return the string corn."""
    return "corn"



# !/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Setup module for setuputils-esque tools that support it.

I don't really know how these work.
"""

import setuptools

with open("README.rst", "r") as fd:
    long_description = fd.read()

with open("LICENSE") as fd:
    license = fd.read()

setuptools.setup(
    name="quizmake",
    version="0.0.8.3",
    author="Jason Nguyen",
    author_email="jnguye21@uoguelph.ca",
    description="Question generator package for D2L and Moodle",
    install_requires=["wheel", "pyparsing"],
    license="MIT",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    url="https://github.com/jnguyen1098/quizmake",
    packages=setuptools.find_packages(),
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Environment :: Console",
        "Intended Audience :: Education",
        "Natural Language :: English",
        "Operating System :: POSIX",
        "Operating System :: Unix",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: MacOS",
        "Programming Language :: Python :: 3",
        "Topic :: Education :: Testing",
        "License :: OSI Approved :: MIT License",
    ],
    keywords="moodle desire2learn d2l",
    python_requires=">=3.7",
)



# !/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Grammar module for quizmake's question format(s).

This was mostly an exercise in using pyparsing.
"""

from pyparsing import (
    CaselessLiteral,
    Group,
    Literal,
    OneOrMore,
    ParserElement,
    Optional,
    ParseResults,
    Regex,
    restOfLine,
    ZeroOrMore,
)


def parse_file(filename: str) -> ParseResults:
    """Parse a file using the grammar."""
    return bnf.parseFile(filename)


# This makes sure nothing is whitespace
ParserElement.setDefaultWhitespaceChars("")

"""
Multiple choice question
"""

newline = Literal("\n")

comment_line = Literal("//") + restOfLine + newline

content_line = Regex("[^[].*") + newline.suppress()

question_header = CaselessLiteral("[multiple_choice]") + newline.suppress()
question_section = question_header + OneOrMore(content_line)

answer_header = CaselessLiteral("[answer]") + newline.suppress()
answer_section = answer_header + OneOrMore(content_line)

feedback_header = CaselessLiteral("[feedback]") + newline.suppress()
feedback_section = feedback_header + OneOrMore(content_line)

multiple_choice_q = (
    ZeroOrMore(comment_line).suppress() + Group(question_section) + Group(answer_section) + Group(feedback_section)
)

"""
Short answer question
"""
short_answer_q = Literal("Not implemented")

"""
True-False question
"""
true_false_q = Literal("Not implemented")

"""
Matching question
"""
matching_q = Literal("Not implemented")

"""
Numerical question
"""
numerical_q = Literal("Not implemented")

"""
Essay question (free-form)
"""
essay_q = Literal("Not implemented")

"""
Description (not a question)
"""
description_q = Literal("Not implemented")

"""
Cloze question
"""
cloze_q = Literal("Not implemented")

question = (
    multiple_choice_q
    | short_answer_q
    | true_false_q
    | matching_q
    | numerical_q
    | essay_q
    | description_q
    | cloze_q
)
bnf = question
